/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: in 20 Minutes (Coffee Break Series)
	 Beginners Taste

     Minimal working examples (MWE) to build upon and 
	 modify to gain an understanding of the API

	 Extensions to the Vulkan Graphics Coffee Break Series
	   
/*
/****************************************************************/
/*
	Loads 3D obj model - extracts the model vertex/normals/texture
	information (including material colours) - and constructs
	the Vulkan buffers - all ready for drawing/including
	in the main program (i.e., command buffer).


*/
/****************************************************************/



#pragma once

#include <stdio.h>
#include <stdlib.h>

using namespace std;
#include <vector>
#include <array>

#include "tgatexture.h"

// `tiny_obj_loader' is a minimal self contained loader
// ideal for loading in obj files - handle multiple models
// materials
// https://syoyo.github.io/tinyobjloader/
#include "tiny_obj_loader.h"

// Vulkan handles
extern VkCoffeeDevice *				coffeeDevice;

// About obj file format:
// ref: https://en.wikipedia.org/wiki/Wavefront_.obj_file



class VkCoffeeObjLoader
{
public:
	struct Vertex 
	{
		float pos[3];
		float uv[2];
		float col[3];
		float normal[3];
		int   matId;
	};

	struct VerticeDescription
	{
		VkPipelineVertexInputStateCreateInfo           inputState;
		std::vector<VkVertexInputBindingDescription>   bindingDescriptions;
		std::vector<VkVertexInputAttributeDescription> attributeDescriptions;
	};

	VkBuffer				vbuf  = VK_NULL_HANDLE;
	VkDeviceMemory			vmem  = VK_NULL_HANDLE;
	size_t					vsize = 0;
	
	VkBuffer				ibuf  = VK_NULL_HANDLE;
	VkDeviceMemory			imem  = VK_NULL_HANDLE;
	size_t					isize = 0;

	uint32_t				indexCount;
	Vector3					dim;
	VerticeDescription		verticeDescription;

	vector<VkTGATexture*>	tgaTextures;

	VkCoffeeObjLoader(const char* fileName, const char* mtlFolder)
	{
		LoadMesh(fileName, mtlFolder);

		SetupVertexDescriptions();
	}

	void LoadMesh(const char*      fileName, 
				  const char*      mtlFolder)
	{
		tinyobj::attrib_t attrib;
		std::vector<tinyobj::shape_t> shapes;
		std::vector<tinyobj::material_t> materials;

		std::string err;
		bool ret = tinyobj::LoadObj(&attrib, 
		                            &shapes, 
									&materials, 
									&err, 
									fileName,
									mtlFolder);

		if (!err.empty()) 
		{ // `err` may contain warning message.
			printf("%s \n", err.c_str() );
		}


		printf("Number of shapes:    %d \n", shapes.size());
		printf("Number of materials: %d \n", materials.size());


		std::vector<Vertex> vertexBuffer;

		#if 1
		for (size_t s = 0; s < shapes.size(); s++) 
		{
			for (size_t f = 0; f < shapes[s].mesh.indices.size() / 3; f++) 
			{
				tinyobj::index_t idx0 = shapes[s].mesh.indices[3 * f + 0];
				tinyobj::index_t idx1 = shapes[s].mesh.indices[3 * f + 1];
				tinyobj::index_t idx2 = shapes[s].mesh.indices[3 * f + 2];
        
				int current_material_id = shapes[s].mesh.material_ids[f];

				if ((current_material_id < 0) || (current_material_id >= static_cast<int>(materials.size()))) 
				{
					// Invaid material ID. Use default material.
					current_material_id = materials.size() - 1; // Default material is added to the last item in `materials`.
				}
				DBG_ASSERT(current_material_id>=0);

				float diffuse[3] = {0};
				for (size_t i = 0; i < 3; i++) 
				{
					diffuse[i] = materials[current_material_id].diffuse[i];
				}

				float tc[3][4] = {0};
				if (idx0.texcoord_index > -1 &&
				    idx1.texcoord_index > -1 &&
					idx2.texcoord_index > -1) 
				{
				    DBG_ASSERT((int)attrib.texcoords.size() > 2 * idx0.texcoord_index + 1);
				    DBG_ASSERT((int)attrib.texcoords.size() > 2 * idx1.texcoord_index + 1);
				    DBG_ASSERT((int)attrib.texcoords.size() > 2 * idx2.texcoord_index + 1);
				    tc[0][0] = attrib.texcoords[2 * idx0.texcoord_index];
				    tc[0][1] = attrib.texcoords[2 * idx0.texcoord_index + 1];
				    tc[1][0] = attrib.texcoords[2 * idx1.texcoord_index];
				    tc[1][1] = attrib.texcoords[2 * idx1.texcoord_index + 1];
				    tc[2][0] = attrib.texcoords[2 * idx2.texcoord_index];
				    tc[2][1] = attrib.texcoords[2 * idx2.texcoord_index + 1];
				}


				const int vf0 = idx0.vertex_index;
				const int vf1 = idx1.vertex_index;
				const int vf2 = idx2.vertex_index;
				DBG_ASSERT(vf0 >= 0);
				DBG_ASSERT(vf1 >= 0);
				DBG_ASSERT(vf2 >= 0);
				float v[3][3];
				for (int k = 0; k < 3; k++) 
				{
					v[0][k] = attrib.vertices[3 * vf0 + k];
					v[1][k] = attrib.vertices[3 * vf1 + k];
					v[2][k] = attrib.vertices[3 * vf2 + k];
				}// End for k

				
				float n[3][3] = {0};
				if ( idx0.normal_index > -1 &&
				     idx1.normal_index > -1 &&
					 idx2.normal_index > -1 )
				{
					const int nf0 = idx0.normal_index;
					const int nf1 = idx1.normal_index;
					const int nf2 = idx2.normal_index;
					DBG_ASSERT(nf0 >= 0);
					DBG_ASSERT(nf1 >= 0);
					DBG_ASSERT(nf2 >= 0);
				
					for (int k = 0; k < 3; k++) 
					{
						n[0][k] = attrib.normals[3 * nf0 + k];
						n[1][k] = attrib.normals[3 * nf1 + k];
						n[2][k] = attrib.normals[3 * nf2 + k];
					}// End for k
				}

				Vertex vert;
				for (int a = 0; a<3; ++a)
				{
					vert.pos[0] = v[a][0];
					vert.pos[1] = v[a][1];
					vert.pos[2] = v[a][2];

					vert.col[0] = diffuse[0];
					vert.col[1] = diffuse[1];
					vert.col[2] = diffuse[2];

					vert.normal[0] = n[a][0];
					vert.normal[1] = n[a][1];
					vert.normal[2] = n[a][2];

					// no normals in the model - manually
					// calculate
					if ( idx0.normal_index < 0 )
					{
						Vector3 v0 ( v[0][0], v[0][1], v[0][2] );
						Vector3 v1 ( v[1][0], v[1][1], v[1][2] );
						Vector3 v2 ( v[2][0], v[2][1], v[2][2] );
						Vector3 dir = Vector3::Cross( v0-v1, v0-v2 );
						dir = Vector3::Normalize( dir );
						vert.normal[0] = dir.x;
						vert.normal[1] = dir.y;
						vert.normal[2] = dir.z;
					}

					vert.uv[0] = tc[a][0];
					vert.uv[1] = tc[a][1];

					vert.matId = current_material_id;

					if ( materials[current_material_id].diffuse_texname.length() < 5 )
					{
						vert.matId = 999;
					}

					vertexBuffer.push_back(vert);
				}// End for a

			}// End for f
		}// End for s
		#endif


		DBG_ASSERT(materials.size()<15);
		for (int i=0; i<(int)materials.size(); ++i)
		{
			if ( materials[i].diffuse_texname.length() < 5 )
			{
				tgaTextures.push_back( NULL );
				continue;
			}

			std::string texPathAndName = std::string(mtlFolder) + "/textures/" + materials[i].diffuse_texname;
			VkTGATexture* tga = new VkTGATexture( texPathAndName.c_str(),
												  coffeeDevice->device, 
												  coffeeDevice->physicalDevice );
			tgaTextures.push_back( tga );
		}// End for i

		// Setup indices
		std::vector<uint32_t> indexBuffer;
		for (int i=0; i< (int)vertexBuffer.size() ; ++i )
		{
			indexBuffer.push_back( i );
		}// End for i
		indexCount = indexBuffer.size();

		CreateBuffer(
			coffeeDevice,
			VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			vertexBuffer.size() * sizeof(Vertex),
			vertexBuffer.data(),
			&vbuf,
			&vmem);

		// Setup indices
		CreateBuffer(
			coffeeDevice,
			VK_BUFFER_USAGE_INDEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			indexBuffer.size() * sizeof(uint32_t),
			indexBuffer.data(),
			&ibuf,
			&imem);

	}


	
	void SetupVertexDescriptions()
	{
		// Binding description
		verticeDescription.bindingDescriptions.resize(2);

		VkVertexInputBindingDescription vInputBindDescription0 = {};
		vInputBindDescription0.binding   = VERTEX_BUFFER_BIND_ID;
		vInputBindDescription0.stride    = sizeof(Vertex);
		vInputBindDescription0.inputRate = VK_VERTEX_INPUT_RATE_VERTEX;

		VkVertexInputBindingDescription vInputBindDescription1 = {};
		vInputBindDescription1.binding   = INSTANCE_BUFFER_BIND_ID;
		vInputBindDescription1.stride    = sizeof(float) * 6;
		vInputBindDescription1.inputRate = VK_VERTEX_INPUT_RATE_INSTANCE;

		verticeDescription.bindingDescriptions[0] = vInputBindDescription0;
		verticeDescription.bindingDescriptions[1] = vInputBindDescription1;



		// Attribute descriptions
		verticeDescription.attributeDescriptions.resize(7);

		// Location 0 : Position
		verticeDescription.attributeDescriptions[0].location = 0;
		verticeDescription.attributeDescriptions[0].binding  = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[0].format   = VK_FORMAT_R32G32B32_SFLOAT;
		verticeDescription.attributeDescriptions[0].offset   = 0;

		// Location 1 : Texture coordinates
		verticeDescription.attributeDescriptions[1].location = 1;
		verticeDescription.attributeDescriptions[1].binding  = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[1].format   = VK_FORMAT_R32G32_SFLOAT;
		verticeDescription.attributeDescriptions[1].offset   = sizeof(float) * 3;

		// Location 2 : Color
		verticeDescription.attributeDescriptions[2].location = 2;
		verticeDescription.attributeDescriptions[2].binding  = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[2].format   = VK_FORMAT_R32G32B32_SFLOAT;
		verticeDescription.attributeDescriptions[2].offset   = sizeof(float) * 5;

		// Location 3 : Normal
		verticeDescription.attributeDescriptions[3].location = 3;
		verticeDescription.attributeDescriptions[3].binding  = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[3].format   = VK_FORMAT_R32G32B32_SFLOAT;
		verticeDescription.attributeDescriptions[3].offset   = sizeof(float) * 8;

		// Location 4 : Material Id
		verticeDescription.attributeDescriptions[4].location = 4;
		verticeDescription.attributeDescriptions[4].binding  = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[4].format   = VK_FORMAT_R32_UINT;
		verticeDescription.attributeDescriptions[4].offset   = sizeof(float) * 11;

		// ************************
		// These are fetched for each instance rendered
		// ************************
		// Location 5 : Normal
		verticeDescription.attributeDescriptions[5].location = 5;
		verticeDescription.attributeDescriptions[5].binding  = INSTANCE_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[5].format   = VK_FORMAT_R32G32B32_SFLOAT;
		verticeDescription.attributeDescriptions[5].offset   = sizeof(float) * 0;

		// Location 6 : Material Id
		verticeDescription.attributeDescriptions[6].location = 6;
		verticeDescription.attributeDescriptions[6].binding  = INSTANCE_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[6].format   = VK_FORMAT_R32G32B32_SFLOAT;
		verticeDescription.attributeDescriptions[6].offset   = sizeof(float) * 3;

		




		VkPipelineVertexInputStateCreateInfo pipelineVertexInputStateCreateInfo = {};
		pipelineVertexInputStateCreateInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
		pipelineVertexInputStateCreateInfo.pNext = NULL;

		verticeDescription.inputState                                 = pipelineVertexInputStateCreateInfo;
		verticeDescription.inputState.vertexBindingDescriptionCount   = verticeDescription.bindingDescriptions.size();
		verticeDescription.inputState.pVertexBindingDescriptions      = verticeDescription.bindingDescriptions.data();
		verticeDescription.inputState.vertexAttributeDescriptionCount = verticeDescription.attributeDescriptions.size();
		verticeDescription.inputState.pVertexAttributeDescriptions    = verticeDescription.attributeDescriptions.data();
	}// End SetupVertexDescriptions(..)

	VkBool32 CreateBuffer(VkCoffeeDevice*		coffeeDevice,
	                      VkBufferUsageFlags    usageFlags, 
	                      VkMemoryPropertyFlags memoryPropertyFlags, 
						  VkDeviceSize          size, 
						  void *                data, 
						  VkBuffer *            buffer, 
						  VkDeviceMemory *      memory)
	{
		VkMemoryRequirements memReqs;
		VkMemoryAllocateInfo memAlloc = {};
		memAlloc.sType           = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		memAlloc.pNext           = NULL;
		memAlloc.allocationSize  = 0;
		memAlloc.memoryTypeIndex = 0;

		VkBufferCreateInfo bufferCreateInfo = {};
		bufferCreateInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
		bufferCreateInfo.pNext = NULL;
		bufferCreateInfo.usage = usageFlags;
		bufferCreateInfo.size  = size;
		bufferCreateInfo.flags = 0;


		VK_CHECK_RESULT(vkCreateBuffer(coffeeDevice->device, &bufferCreateInfo, nullptr, buffer));

		vkGetBufferMemoryRequirements(coffeeDevice->device, *buffer, &memReqs);
		memAlloc.allocationSize  = memReqs.size;
		memAlloc.memoryTypeIndex = coffeeDevice->GetMemoryType(memReqs.memoryTypeBits, memoryPropertyFlags);
		VK_CHECK_RESULT(vkAllocateMemory(coffeeDevice->device, &memAlloc, nullptr, memory));
		if (data != nullptr)
		{
			void *mapped;
			VK_CHECK_RESULT(vkMapMemory(coffeeDevice->device, *memory, 0, size, 0, &mapped));
			memcpy(mapped, data, (size_t)size);
			vkUnmapMemory(coffeeDevice->device, *memory);
		}
		VK_CHECK_RESULT(vkBindBufferMemory(coffeeDevice->device, *buffer, *memory, 0));

		return true;
	}// End CreateBuffer(..)

}; // End class VkCoffeeObjLoader


